/****************************************************************
********
*************************  NARROWING  
***************************
*****************************************************************
********
* FICHERO narrow.c: Este fichero contiene los procedimientos que
* implementan las reglas del basic conditional narrowing. Este es
uno de 
* los metodos mas conocidos para lograr la unificacion semantica. 
* Los unicos puntos de entrada de este fichero son las funciones:
* 
*    narrower(); Calculo en anchura de los niveles del arbol de
*         narrowing. Da como resultado un conjunto de estados de
*         exito, que contienen las substituciones que son
respuestas 
*         computadas.
*
*****************************************************************
********/
#define pep "\n Ocurrencias "
#define pep1 "\n ecuacion estado "
#define pep2 "\n ecuacion teoria "
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"
                                 
extern void continua_ocu(struct ocus *locu,struct lista *ocu_sub, 

		     struct lista **lfin,struct lista **lini, char *vb);




/* Prototipos de funciones externas, definidas en Procesos
auxiliares. No      
todas estas funciones son empleadas en "narrow.c". */ 
#include "proaux.h"

/* Prototipos de funciones externas, definidas en el modulo de
Unificacion. No      
todas estas funciones son empleadas en "narrow.c". */ 
#include "unific.h"



/**********************   GESTION DE LA MEMORIA   
**********************/


/*elimina toda la memoria asociada a una sustitucion, pero
teniendo cuidado de no borrar la lista de ecuaciones original
apuntada por el campo raiz. Esta es otra chapuza para propiciar
la eficiencia del proceso de narrowing. Nos evitariamos todo
esto si no hubiera que guardar informacion de la ecuaciones
originales, en terminos de un puntero a las ecuaciones, para el
calculo de la semantica operacional.
Lo de NR es para indicar que No borra las ecuaciones Raiz. */

PRIVATE void borrastatNR (struct stat *est)
{ if (est) {
     borrastatNR(est->sig);
     borralis(est->ecus);
     borrasub(est->alfa);
     free(est);};
}


/**************************   OCURRENCIAS  
*****************************/

/* Produce una copia, dejando inalterada la lista de ecuaciones
original. En la nueva 
lista una de las ecuaciones, "vieja", se reemplaza por otra,
"nueva". La ecuacion "nueva"
no tiene que replicarse, simplemente se enlaza a la estructura
lista. La ecuacion
vieja si es borrada. La funcin retorna un puntero a la lista
creada. */

PRIVATE struct lista *reemplazar(struct lista *ini, struct arbol
*vieja, struct arbol *nueva)
{
  struct lista *listaux=NULL, *listaux1=NULL, *listaux2=NULL,
    *listaux3=NULL;
  for (listaux1=ini; listaux1; listaux1=listaux1->sig) {
    listaux2=noulista();
    if (listaux1->ecu==vieja) listaux2->ecu=nueva;
    else listaux2->ecu=copiecu(listaux1->ecu);
    listaux2->sig=NULL;
    if (listaux == NULL)  listaux=listaux2;
    else  listaux3->sig=listaux2; 
    listaux3=listaux2;
  }
  return listaux;
}


/* Extrae todas las ecuaciones del cuerpo de una regla y las
incluye
en la lista de ecuaciones que se pasa como primer parmetro. */ 

PRIVATE void incluye(struct lista *ini, struct arbol *body)
{
     struct lista *listaux1=NULL;

     for (listaux1=ini; listaux1->sig; listaux1=listaux1->sig);
     for (;body;)
       { if ((listaux1->sig=(struct lista *) malloc(sizeof(struct
lista))) == NULL)
                    memerror();
         listaux1=listaux1->sig;
         listaux1->ecu=body;
         body=body->herm->herm;
         listaux1->ecu->herm->herm=NULL;
       }
     listaux1->sig=NULL;
}         


/************************* REGLAS DE BASIC NARROWING
************************/
/****************************************************************
*************/


/* Aplicacion de la primera regla de "Basic Narrowing".
A BN1() se le llamara suministrandole un estado de narrowing.
Devolvera NULL en
caso de error y un puntero a una estructura estado en caso de
exito. El campo 
"ecus" de esta estructura apuntara siempre a NULL; el campo
"raiz" a la lista 
de ecuaciones de partida;  el campo "alfa" a una substitucion; y
el puntero a 
siguiente "sig" quedara inicializado a NULL.

Esta funcion es la  que completa el proceso de unificacion
sintactica. Se
apoya en la funcion unificador() para tratar la unificacion de un
conjunto de 
ecuaciones. Sin embargo es algo mas que un simple proceso de
unificacion
sintactica, pues  comienza calculando el mgu de una instancia
G.theta de un
conjunto de ecuaciones y termina componiendo theta con el mgu
obtenido. Ahora
bien, cuando theta==NULL, estamos en el caso de una unificacion
sintactica pura y llana. */

PRIVATE struct stat *BN1(struct stat *estado)
{
     struct arbol *ecuaux, *izq, *der;
     struct lista *listaux;
     struct subs *mgu, *subaux;
     struct stat *estaux;

     st.BN_RU1_TECAt++; mgu = NULL; subaux = NULL;
     for (listaux = estado->ecus; listaux; listaux=listaux->sig)
{
          ecuaux = copiecu(listaux->ecu);
          aplicar(&ecuaux, estado->alfa);
          aplicar(&ecuaux, mgu);
          der= ecuaux->herm; ecuaux->herm= NULL; izq= ecuaux;
          if (unificador(izq, der, &subaux)) {
               borraecu(izq); izq=NULL;
               borraecu(der); der=NULL;
               mgu = componer(mgu, subaux);
               borrasub(subaux); subaux=NULL;
          }
          else {
                  borraecu(izq); borraecu(der);
               borrasub(mgu);
               return (struct stat *) NULL;
          }
     }
     
     if ( (estaux= (struct stat *) malloc(sizeof(struct stat)))
== NULL)
                         memerror();
     estaux->ecus= NULL;
     estaux->raiz= estado->raiz;
     estaux->alfa= componer(estado->alfa, mgu);
     estaux->sig= NULL;
     borrasub(mgu);
     st.BN_RU1_TECAe++;
     return  estaux;
}


/* Aplicacion de la segunda regla de "Basic Narrowing". 
A BN2() se le suministrara el estado que se analiza en ese
momento, una ecuacion
del objetivo de ese estado (la que habria que reemplazar en caso
de exito al 
encontrar una ocurrencia que se pudiese unificar), una ocurrencia
de la ecuacion
, y una regla de la teoria, estandarizada, con la que unificar su
parte 
izquierda.
En caso de fracaso devolvera NULL. En caso de exito devolvera un
puntero a un 
estado de narrowing formado por: un puntero a una lista de
ecuaciones en la que
la ecuacion, "ecu", aparecera reescrita; un puntero a la lista de
ecuaciones 
de partida; un puntero a una substitucion; y el puntero a
siguiente inicializado 
a NULL.  */

PRIVATE struct stat *BN2(struct stat *estado, struct arbol *ecu,
struct
arbol *ocu, struct arbol *reglateo)
{
     struct stat *estaux=NULL;
     struct arbol *eu=NULL, *landa=NULL, *ro=NULL, *euro=NULL,
*body=NULL;
     struct subs *subaux=NULL;

     st.BN_RU2t++;
     /* Hay que comprobar si la parte derecha, landa, de la
cabeza de la regla
     estandarizada se unifica sintacticamente con la instancia de
la
     ocurrencia, ocu, resultado de aplicar la sustitucion
estado->alfa. */
     eu=copiterm(ocu);
     aplicar(&eu,estado->alfa);
     landa=copiterm(reglateo);
     subaux=NULL;
     if(trace){
          printf("\n abans de unificador\n");
          verecu(eu);
          verecu(landa);
     }
     if (unificador(eu,landa,&subaux)) {
          if (trace) {printf("\n sustitucions:");
                      verunif(subaux);}
          st.BN_RU2e++;
          if ((estaux=(struct stat *)malloc(sizeof(struct
stat)))==NULL)
                    memerror();
          ro=copiterm(reglateo->herm);
          body=copiecu(reglateo->herm->herm);
          /*  reglateo->herm->herm=NULL;    */
          euro=modifica(ecu, ocu, ro);
          estaux->ecus=reemplazar(estado->ecus, ecu, euro);
          if (body) incluye(estaux->ecus, body);
          estaux->raiz=estado->raiz;
          estaux->alfa=componer(estado->alfa,subaux);
          estaux->sig=NULL;
     }
     borraecu(eu); borraecu(landa);
borraecu(ro);borrasub(subaux);
     return estaux;
}


/****************************************************************
**************/
/***************************** CALCULO EN ANCHURA
*****************************/

 /* El procedimiento narrower()  generara el arbol de narrowing,
todos los 
niveles, que obtiene uno a uno en un calculo en anchura. Tiene
como argumento 
un conjunto de estados de narrowing. Cuando se aplique narrower()
a la 
resolucion de un objetivo, G::= <=e1,..,eN, la lista de partida
estara formada
por un solo estado <G,  (e1,..,eN), Ident>. Devuelve una lista de
estados de 
exito, <<=true, raiz, theta>, que se derivan del conjunto de
estados inicial.
Por cada nivel de arbol, tendremos que realizar los siguientes
pasos:

A)   Aplicar BN1() a cada estado de ese nivel. Los exitos
obtenidos se 
enlazaran en la lista de respuesta, R.
B)   Aplicar BN2(), para cada estado del nivel, para cada
ecuacion de la 
lista de ecuaciones del estado, para cada ocurrencia de esa
ecuacion, para cada
regla estandarizada de la teoria. Los exitos de BN2() se enlazan
en una lista
de estados que formara el siguiente nivel del arbol de narrowing
a analizar.
C)   El algoritmo terminara cuando la lista de estados del nivel
alcanzado 
sea vacia, es decir, el apuntador que la referencia sea NULL.

Si el proceso de narrowing no tiene exito el proceso terminara
cuando se agote 
la memoria. Un procedimiento mas correcto para controlar este
tipo de 
situaciones podria ser aplicar un metodo heuristico, en el que
solo se 
permitiese al algoritmo calcular un numero N de niveles del
arbol. Este numero 
vendria dictado por la experiencia (Por el momento no nos
plantearemos este 
problema en la presente version). */


PUBLIC struct stat *narrower1(struct stat *nivel)
{
  struct arbol *stdet; /* puntero a copia estandarizada de una
regla 
               de la teoria. */
  struct lista      *g, /* puntero a lista de ecuaciones
"objetivo" de un estado.*/
          *u, /* puntero a cabeza lista de ocurrencias no
variables de
                 una ecuacion de g */
          *ufin, /* Idem. al final de la lista */
          *et, *uaux; /* variables auxiliares */
  struct stat  *L=NULL, /*puntero a nivel de estados de narrowing
que se va 
                a analizar. */
          *L1=NULL, /* Idem. Se utiliza como puntero auxiliar
para 
                 recorrer el nivel. */
          *L2=NULL, /* puntero a lista que almacena estados de
narrowing
                  G2. Siguiente nivel a analizar */
          *L2fin=NULL, /* puntero al final de la lista L2. */
          *G1=NULL, /* puntero a estados de exito de BN1(). */
          *G2=NULL, /* puntero a estados de exito de BN2(). */
          *R=NULL, /* puntero lista que almacena estados de exito
G1. 
                   Es la respuesta devuelta. */
          *Rfin=NULL; /* puntero al final de la lista Rfin. */
  struct stat *Lini; /* para chapuza */

  if (nivel == NULL) return (struct stat *) NULL; /* Para que
seguir. */

  L= (struct stat *) copistat(nivel); /* copia lista de estados
de partida, que
                         este procedimiento narrower() deja
                         inalterada. */

  /* Esta linea es la chapuza necesaria para evitarnos que BN1()
y BN2()
  tengan que copiar las ecuaciones raiz cada vez que devuelven un
estado. BN1() 
  y BN2() se han implementado de forma que solamente apuntan a
las ecuaciones
  raiz; esto nos obliga a no borrar el nivel de estados de
narrowing del que
  partimos. Primero tomamos nota de donde empieza y donde
termina. Tambien nos
  obliga a utilizar borrastatNR(), en lugar de borrastat(). */
  Lini = L;

  while(L!=NULL) { /* Mientras haya niveles que analizar */
     for (L1=L,L2= NULL; L1; L1=L1->sig) 
     { /* Mientras nivel de estados de narrowing no vacio,
          para cada estado de L1 */
          G1= BN1(L1);
          if (G1!= NULL) { /* anyadir G1 a lista de estados de
exito R. */
               if (R==NULL) R=G1;
               else Rfin->sig=G1;
               Rfin=G1;
               if (trace) {
                    printf("\nNARROWER : ESTADO DE EXITO:\n");
                    verestado(G1);
               };
          };
          for (g=L1->ecus; g; g=g->sig)
          { /* Para cada ecuacion de la lista de ecus del estadoL1 */
            ufin=NULL;
            u= ocubas(&(g->ecu), &ufin); /* Calcula ocurrencias no var de una ec.*/
            for (uaux= u; uaux; uaux= uaux->sig)
            { /* Para cada ocurrencia de la ecuacion objetivo g->ecu */
               for(et=listaini; et; et= et->sig)
               { /* Para cada ecuacion de la teoria */
                 stdet= estandar(et->ecu); /* estandariza regla teoria.*/
                 if(trace){
                         printf("\n abans de BN2\n");
                         verecu(uaux->ecu);
                         verecu(stdet);
                              }
                 G2= BN2(L1, g->ecu, uaux->ecu, stdet);
                 if (G2 != NULL) { /* anyadir G2 a lista de ests L2.*/
                   if (L2==NULL) L2=G2;
                   else L2fin->sig=G2;
                   L2fin=G2;
                   if (trace) {
                    printf("\nNARROWER : ESTADO DE NARROWING:\n");
                    verestado(G2);
                   };
                 };
                 borraecu(stdet);  
               };
            };
            borraocu(u); /* libera espacio lista de ocurrencias*/
          };
     };
     if (L != Lini)  borrastatNR(L);
     L=L2; /* siguiente nivel a analizar */
     if (trace) {
     printf("NARROWER : lista de estados de narrowing, siguiente nivel.\n");
     verestado(L);
     };
  };
  /* Se copian las ecuaciones originales de la lista de exitos,
que hasta
  ahora solo estaban apuntadas. Supongo que esta informacin no
se ha perdido
  en los sucesivos borrados de listas de estados que se han
realizado, ya que
  cada L1->raiz debe apuntar a una lista de ecuaciones  raiz de
la lista de
  estados original, Lini, que se ha tenido cuidado de no borrar
hasta el
  momento. */
  if (R!= NULL) for (L1=R; L1; L1=L1->sig)
L1->raiz=copilis(L1->raiz);
  borrastat(Lini); /* Ahora ya se puede borrar la copia de la
lista de
             estados original. Ya se puede borrar las listas de
             ecuaciones raiz. */
  if (trace) {
     printf("NARROWER : lista de estados de exito R.\n");
     verestado(R);
  };

 return R;
}


/*                    NARROWING GENERAL                  */

/* Devuelve un int con 1 si no hay ocurrencia reescribibles, 0 en el otro caso */

PRIVATE int no_res(struct lista *ocus,struct funcs *ocueq)
{
  struct funcs *aux;
  if (ocus==NULL) return 1;
  for (;ocus!=NULL;ocus=ocus->sig)
    for(aux=ocueq;aux!=NULL;aux=aux->sig)
      {if (!strcmp(aux->nombre,ocus->ecu->oc)) return 0;}

  return 1;
}
	
	/* el entero ret_exit indica si se deben calcular o no todas las 
	soluciones, 1 (se sale en cuanto encuentra una solucion), 0 (las 
	calcula todas*/
	
PUBLIC struct stat *narrower(struct stat *nivel,struct lista *listeq,
				struct funcs *ocueq,int ret_exit)
{
  struct arbol *stdet,*euro,*izq,
    *u1=NULL,*u2=NULL; /* puntero a copia estandarizada de una
			  regla de la teoria. */
  struct lista      *g,*gaux, /* puntero a lista de ecuaciones
			   "objetivo" de un estado.*/
    *u, /* puntero a cabeza lista de ocurrencias no
	   variables de
	   una ecuacion de g */
    *ufin, /* Idem. al final de la lista */
    *et, *uaux, /* variables auxiliares */
    *laux;
  struct stat  *L=NULL, /*puntero a nivel de estados de narrowing
			  que se va 
			  a analizar. */
    *L1=NULL, /* Idem. Se utiliza como puntero auxiliar
		 para 
                 recorrer el nivel. */
    *L2=NULL, /* puntero a lista que almacena estados de
		 narrowing
		 G2. Siguiente nivel a analizar */
    *L2fin=NULL, /* puntero al final de la lista L2. */
    *G1=NULL, /* puntero a estados de exito de BN1(). */
    *G2=NULL, /* puntero a estados de exito de BN2(). */
    /*    *G2t=NULL,*G1t=NULL,*/
    *R=NULL, /* puntero lista que almacena estados de exito
		G1. 
		Es la respuesta devuelta. */
    *Rfin=NULL; /* puntero al final de la lista Rfin. */
  int k=0,cont=0;
  struct subs  *mgu=NULL;
  
  if (nivel == NULL) return (struct stat *) NULL; /* Para que
						     seguir. */
  
  L= (struct stat *) copistat(nivel); /* copia lista de estados
					 de partida, que
					 este procedimiento narrower() deja
                         inalterada. */
  
  /* Mientras haya niveles que analizar  y no nos pasemos de la 
  profundidad limite, indicada por el argumento prof*/
  
  while((L!=NULL)&&(cont<prof)) { 
  cont++;
  
    for (L1=L,L2= NULL; L1; L1=L1->sig) 
      { /* Mientras nivel de estados de narrowing no vacio,
	   para cada estado de L1 */
        for (g=L1->ecus; g; g=g->sig)
          { /* Para cada ecuacion de la lista de ecus del estado
	       L1 */
	    if (trace) {
	      printf(pep1);
	      verecu(g->ecu);
            }
	    
	    mgu=NULL;
	    izq=copiterm(g->ecu); 
	    k=0;
	    if (unificador(izq,g->ecu->herm,&mgu))
	      {
		k=1;
		/*si solo existe esta ecuacion, se trata de un exito*/
		if ((g->sig==NULL)&&(g==L1->ecus)){
		  if ( (G1= (struct stat *) malloc(sizeof(struct stat))) == NULL)
		    memerror();
		  /*	G1t->ecus=reemplazar(L1->ecus,g->ecu,euro);*/
		  G1->ecus=copilis(g);
		  G1->raiz=NULL;
		  G1->alfa=componer(L1->alfa,mgu);
		  G1->sig=NULL;
		  
		  if (R==NULL) R=G1;
		  else Rfin->sig=G1;
		  Rfin=G1;
		  if (trace) {
		    printf("\nNARROWER ESTADO DE EXITO:\n");
		    verestado(G1);
		  };
		  if (ret_exit) {
		  	/* como ya hemos encontrado solucion, nos salimos*/
		  	borrasub(mgu);
		        borraecu(izq);
		        if (L != NULL)  borrastat(L);
		  	return R;
		  	}
		}
	      }
	
	    borrasub(mgu);
	    borraecu(izq);
            ufin=NULL;
            u= ocubas(&(g->ecu), &ufin); /* Calcula ocurrencias
					    no var de una ec.*/
	    if (trace) {  
	      verocu(pep,u);
	    		}
	    /*comprobar que no tengan ocurrencias reescribibles*/ 
	    if (no_res(u,ocueq))
	      {
		/*si ha habido exito se elimina la ecuacion*/
		if (k==1) L1->ecus=L1->ecus->sig;
		else break;
	      }
	    else
	      {
		for (uaux= u; uaux; uaux= uaux->sig)
		  { /* Para cada ocurrencia de la ecuacion objetivo
		       g->ecu */
		    for(et=listeq; et; et= et->sig)
		      { /* Para cada ecuacion de la teoria */
			if (trace) {
			  printf(pep2);
			  verecu(et->ecu);
			  
			  }
			
			stdet= estandar(et->ecu); /* estandariza regla teoria.*/
			u1=copiterm(uaux->ecu);
			u2=copiterm(stdet);
			if (trace) {
			  printf("\n ocurrencia");
			  verecu(u1);
			  } 
			mgu=NULL;
			if (unificador(u1,u2,&mgu)){
			  if (trace) {
			    printf("\n sustitucions:");
			    verunif(mgu);
			  }
			  /*cambiar la ocurrencia por parte derecha*/
			  
			  euro=modifica(g->ecu,uaux->ecu,stdet->herm);
			  /*if (trace) {
			    printf("\necuacions\n");
			    verecu(g->ecu);
			    verterm(uaux->ecu);
			    verecu(stdet->herm);
			    verecu(euro);
			    };*/
			  laux=reemplazar(L1->ecus,g->ecu,euro);
			  for (gaux=laux;gaux;gaux=gaux->sig)
			    aplicar(&gaux->ecu,mgu);
			  
			  if ( (G2= (struct stat *)
			    malloc(sizeof(struct stat))) == NULL)
			    memerror();
			    	  
			  G2->ecus=copilis(laux);
			  G2->raiz=NULL;
			  G2->alfa=componer(L1->alfa,mgu);
			  G2->sig=NULL;
			  if (L2==NULL) L2=G2;
			  else L2fin->sig=G2;
			  L2fin=G2;
			  if (trace) {
			    printf("\nNARROWER : ESTADO DE NARROWING:\n");
			    verestado(G2);
			  };
			  borralis(laux); /* borra laux y euro*/
			};/*if*/
			borraecu(u1);borraecu(u2);borrasub(mgu);borraecu(stdet);
		      };/*for*/
		  };/*for*/
            borraocu(u); /* libera espacio lista de ocurrencias*/
	      }/*else*/
          };/*for*/
      };/*for*/
    if (L != NULL)  borrastat(L);
    L=L2; /* siguiente nivel a analizar */
    if (trace) {
      printf("\nNARROWER : lista de estados de narrowing siguiente nivel.");
      verestado(L);
    };
  };/*while*/
  
  if (L != NULL)  borrastat(L); /* por si hemos salido por limite de prof*/
  
  /* Se copian las ecuaciones originales de la lista de exitos,
que hasta  ahora solo estaban apuntadas. Supongo que esta informacin no
se ha perdido en los sucesivos borrados de listas de estados que se han
realizado, ya que cada L1->raiz debe apuntar a una lista de ecuaciones  raiz de
la lista de estados original, Lini, que se ha tenido cuidado de no borrar
hasta el momento. */
/*  if (R!= NULL) for (L1=R; L1; L1=L1->sig)
    L1->raiz=copilis(L1->raiz);*/
  if (trace) {
    printf("NARROWER : lista de estados de exito R.\n");
    verestado(R);
  };
  
  return R;
}

